from __future__ import annotations  # NOTE: prevents the use of < python3.7

from enum import Enum
from functools import reduce
from math import log2
from math import log10


class Unit:
    def __init__(self, name, *aliases):
        self.name = name
        self.aliases = aliases

    def __repr__(self):
        return f'<{self.name}>'


class Units(Enum):
    none = Unit('')
    volt = Unit('volt', 'v', 'V')  # NOTE: a "volt" is a derived unit; treat accordingly in future
    liter = Unit('liter', 'l', 'L')         # volume
    gram = Unit('gram', 'g', 'G')           # mass
    second = Unit('second', 's', 'secs')    # time
    kelvin = Unit('kelvin', 'K')            # temperature
    mole = Unit('mole', 'mol')              # amount
    coulomb = Unit('coulomb', 'C')          # charge

    def __truediv__(self, another_unit) -> Unit:
        if self.name == another_unit.name:
            return Units.none
        elif self == Units.none:
            return another_unit
        elif another_unit == Units.none:
            return self
        else:
            raise ValueError(f'Unsupported unit division: {self.name} / {another_unit.name}. Consider using Ratio.')

    @staticmethod
    def all():
        return dict(list(map(lambda x: (x.value, x.value.aliases), Units)))


# TODO -- consider using a dataclass
class Exponent:
    bases ={2: log2, 10: log10}

    def __init__(self, base: int, power: int):
        if base not in self.bases.keys():
            raise ValueError(f'Only the following bases are supported: {reduce(lambda a,b: f"{a}, {b}", self.bases.keys())}')
        self.base = base
        self.power = power
        self.evaluated = base ** power

    def parts(self):
        return self.base, self.power

    def __truediv__(self, another_exponent):
        return self.evaluated / another_exponent.evaluated

    def __lt__(self, another_exponent):
        return self.evaluated < another_exponent.evaluated

    def __gt__(self, another_exponent):
        return self.evaluated > another_exponent.evaluated

    def __eq__(self, another_exponent):
        return self.evaluated == another_exponent.evaluated

    def __repr__(self):
        return f'<{self.base}^{self.power}>'


class Scale(Enum):
    mebi  = Exponent(2, 20)
    kibi  = Exponent(2, 10)
    mega  = Exponent(10, 6)
    kilo  = Exponent(10, 3)
    hecto = Exponent(10, 2)
    deca  = Exponent(10, 1)
    one   = Exponent(10, 0)
    deci  = Exponent(10,-1)
    centi = Exponent(10,-2)
    milli = Exponent(10,-3)
    micro = Exponent(10,-6)
    _kibi = Exponent(2,-10)
    _mebi = Exponent(2,-20)

    @staticmethod
    def all():
        return dict(map(lambda x: ((x.value.base, x.value.power), x.name), Scale))

    @staticmethod
    def by_value():
        return dict(map(lambda x: (x.value.evaluated, x.name), Scale))

    def __truediv__(self, another_scale):
        power_diff = self.value.power - another_scale.value.power
        if self.value == another_scale.value:
            return Scale.one
        if self.value.base == another_scale.value.base:
            return Scale[Scale.all()[Exponent(self.value.base, power_diff).parts()]]

        base_quotient = self.value.base / another_scale.value.base
        exp_quotient = round((base_quotient ** another_scale.value.power) * (self.value.base ** power_diff), 15)

        if Scale.one in [self, another_scale]:
            power = Exponent.bases[2](exp_quotient)
            return Scale[Scale.all()[Exponent(2, power).parts()]]
        else:
            scale_exp_values = [Scale[Scale.all()[pair]].value.evaluated for pair in Scale.all().keys()]
            closest_val = min(scale_exp_values, key=lambda val: abs(val - exp_quotient))
            return Scale[Scale.by_value()[closest_val]]

    def __lt__(self, another_scale):
        return self.value < another_scale.value

    def __gt__(self, another_scale):
        return self.value > another_scale.value

    def __eq__(self, another_scale):
        return self.value == another_scale.value


# TODO -- consider using a dataclass
class Number:
    def __init__(self, unit: Unit = Units.none, scale: Scale = Scale.one, quantity = 1):
        self.unit = unit
        self.scale = scale
        self.quantity = quantity
        self.value = round(self.quantity * self.scale.value.evaluated, 15)

    def simplify(self):
        return Number(unit=self.unit, quantity=self.value)

    def to(self, new_scale: Scale):
        new_quantity = self.quantity / new_scale.value.evaluated
        return Number(unit=self.unit, scale=new_scale, quantity=new_quantity)

    def as_ratio(self):
        return Ratio(self)

    def __mul__(self, another_number):
        new_quantity = self.quantity * another_number.quantity
        return Number(unit=self.unit, scale=self.scale, quantity=new_quantity)

    def __truediv__(self, another_number) -> Number:
        unit = self.unit / another_number.unit
        scale = self.scale / another_number.scale
        quantity = self.quantity / another_number.quantity
        return Number(unit, scale, quantity)

    def __eq__(self, another_number):
        if isinstance(another_number, Number):
            return (self.unit == another_number.unit) and \
                   (self.quantity == another_number.quantity) and \
                   (self.value == another_number.value)
        elif isinstance(another_number, Ratio):
            return self == another_number.evaluate()
        else:
            raise ValueError(f'"{another_number}" is not a Number or Ratio. Comparison not possible.')

    def __repr__(self):
        return f'<{self.quantity} {"" if self.scale.name == "one" else self.scale.name}{self.unit.value.name}>'


# TODO -- consider using a dataclass
class Ratio:
    def __init__(self, numerator: Number = Number(), denominator: Number = Number()):
        self.numerator = numerator
        self.denominator = denominator

    def reciprocal(self) -> Ratio:
        return Ratio(numerator=self.denominator, denominator=self.numerator)

    def evaluate(self) -> Number:
        return self.numerator / self.denominator

    def __mul__(self, another_ratio):
        new_numerator = self.numerator / another_ratio.denominator
        new_denominator = self.denominator / another_ratio.numerator
        return Ratio(numerator=new_numerator, denominator=new_denominator)

    def __eq__(self, another_ratio):
        if isinstance(another_ratio, Ratio):
            return self.evaluate() == another_ratio.evaluate()
        elif isinstance(another_ratio, Number):
            return self.evaluate() == another_ratio
        else:
            raise ValueError(f'"{another_ratio}" is not a Ratio or Number. Comparison not possible.')

    def __repr__(self):
        # TODO -- resolve int/float inconsistency
        return f'{self.evaluate()}' if self.numerator == self.denominator else f'{self.numerator} / {self.denominator}'




import nox

from os import environ as envvar


PROJECT_NAME = 'ucon'
VENV = f'{PROJECT_NAME}-venv'
TESTDIR = '.'
USEVENV = envvar.get('USEVENV', False)

external = False if USEVENV else True
supported_python_versions = [
    '3.6',
    '3.7',
    '3.8',
    '3.9',
    '3.10',
]

nox.options.default_venv_backend = 'none' if not USEVENV else USEVENV


@nox.session(name=VENV if USEVENV else PROJECT_NAME, python=supported_python_versions)
def tests(session):
    session.run(
        'python', '-m',
        'pip', '--disable-pip-version-check', 'install', '.',
        external=external
    )
    session.run(
        'python', '-m',
        'pip', '--disable-pip-version-check', 'install', '-r', 'requirements.txt',
        external=external
    )
    session.run(
        'python', '-m',
        'coverage', 'run', '--branch',
        '--omit', '.nox/*,noxfile.py,setup.py,test*',
        '--source', '.',
        '-m', 'unittest', 'discover',
        '-s', TESTDIR,
        external=external
    )
    session.run('coverage', 'report', '-m', external=external)
    session.run('coverage', 'xml', external=external)


@nox.session(name=f'build-{PROJECT_NAME}')
def build(session):
    session.run('python', 'setup.py', 'sdist')


# TODO (withtwoemms) -- leverage gitpython to show version on-demand




from unittest import TestCase

from ucon import Number
from ucon import Exponent
from ucon import Ratio
from ucon import Scale
from ucon import Unit
from ucon import Units


class TestUnit(TestCase):

    unit_name = 'second'
    unit_aliases = ('seconds', 'secs', 's', 'S')
    unit = Unit(unit_name, *unit_aliases)

    def test___repr__(self):
        self.assertEqual(f'<{self.unit_name}>', str(self.unit))


class TestUnits(TestCase):

    def test_has_expected_basic_units(self):
        expected_basic_units = {'none', 'volt', 'liter', 'gram', 'second', 'kelvin', 'mole', 'coulomb'}
        self.assertEqual(set(item.name for item in Units), expected_basic_units)

    def test___truediv__(self):
        self.assertEqual(Units.none, Units.gram / Units.gram)
        self.assertEqual(Units.gram, Units.gram / Units.none)
        self.assertEqual(Units.gram, Units.none / Units.gram)

        with self.assertRaises(ValueError):
            Units.gram / Units.liter

    def test_all(self):
        for unit in Units:
            self.assertIsInstance(unit.value, Unit)
        self.assertIsInstance(Units.all(), dict)


class TestExponent(TestCase):

    thousand = Exponent(10, 3)
    thousandth = Exponent(10, -3)

    def test___init__(self):
        with self.assertRaises(ValueError):
            Exponent(5, 3)  # no support for base 5 logarithms

    def test_parts(self):
        self.assertEqual((10, 3), self.thousand.parts())
        self.assertEqual((10, -3), self.thousandth.parts())

    def test___truediv__(self):
        self.assertEqual(1000, self.thousand.evaluated)
        self.assertEqual(float(1/1000), self.thousandth.evaluated)
        self.assertEqual(float(1000000), (self.thousand / self.thousandth))

    def test___lt__(self):
        self.assertLess(self.thousandth, self.thousand)

    def test___gt__(self):
        self.assertGreater(self.thousand, self.thousandth)

    def test___repr__(self):
        self.assertEqual(str(self.thousand), '<10^3>')
        self.assertEqual(str(self.thousandth), '<10^-3>')


class TestScale(TestCase):

    def test___truediv__(self):
        self.assertEqual(Scale.deca, Scale.one / Scale.deci)
        self.assertEqual(Scale.deci, Scale.one / Scale.deca)
        self.assertEqual(Scale.kibi, Scale.mebi / Scale.kibi)
        self.assertEqual(Scale.milli, Scale.one / Scale.deca / Scale.deca / Scale.deca)
        self.assertEqual(Scale.deca, Scale.kilo / Scale.hecto)
        self.assertEqual(Scale._kibi, Scale.one / Scale.kibi)
        self.assertEqual(Scale.kibi, Scale.kibi / Scale.one)
        self.assertEqual(Scale.one, Scale.kibi / Scale.kibi)
        self.assertEqual(Scale.one, Scale.kibi / Scale.kilo)

    def test___lt__(self):
        self.assertLess(Scale.one, Scale.kilo)

    def test___gt__(self):
        self.assertGreater(Scale.kilo, Scale.one)

    def test_all(self):
        for scale in Scale:
            self.assertTrue(isinstance(scale.value, Exponent))
        self.assertIsInstance(Scale.all(), dict)


class TestNumber(TestCase):

    number = Number(unit=Units.gram, quantity=1)

    def test_as_ratio(self):
        ratio = self.number.as_ratio()
        self.assertIsInstance(ratio, Ratio)
        self.assertEqual(ratio.numerator, self.number)
        self.assertEqual(ratio.denominator, Number())

    def test_simplify(self):
        ten_decagrams = Number(unit=Units.gram, scale=Scale.deca, quantity=10)
        point_one_decagrams = Number(unit=Units.gram, scale=Scale.deca, quantity=0.1)
        two_kibigrams = Number(unit=Units.gram, scale=Scale.kibi, quantity=2)

        self.assertEqual(Number(unit=Units.gram, quantity=100), ten_decagrams.simplify())
        self.assertEqual(Number(unit=Units.gram, quantity=1), point_one_decagrams.simplify())
        self.assertEqual(Number(unit=Units.gram, quantity=2048), two_kibigrams.simplify())

    def test_to(self):
        thousandth_of_a_kilogram = Number(unit=Units.gram, scale=Scale.kilo, quantity=0.001)
        thousand_milligrams = Number(unit=Units.gram, scale=Scale.milli, quantity=1000)
        kibigram_fraction = Number(unit=Units.gram, scale=Scale.kibi, quantity=0.0009765625)

        self.assertEqual(thousandth_of_a_kilogram, self.number.to(Scale.kilo))
        self.assertEqual(thousand_milligrams, self.number.to(Scale.milli))
        self.assertEqual(kibigram_fraction, self.number.to(Scale.kibi))

    def test___repr__(self):
        self.assertIn(str(self.number.quantity), str(self.number))
        self.assertIn(str(self.number.scale.value.evaluated), str(self.number))
        self.assertIn(self.number.unit.name, str(self.number))

    def test___truediv__(self):
        some_number = Number(unit=Units.gram, scale=Scale.deca, quantity=10)
        another_number = Number(unit=Units.gram, scale=Scale.milli, quantity=10)
        that_number = Number(unit=Units.gram, scale=Scale.kibi, quantity=10)

        some_quotient = self.number / some_number
        another_quotient = self.number / another_number
        that_quotient = self.number / that_number

        self.assertEqual(some_quotient.value, 0.01)
        self.assertEqual(another_quotient.value, 100.0)
        self.assertEqual(that_quotient.value, 0.00009765625)

    def test___eq__(self):
        self.assertEqual(self.number, Ratio(self.number))  # 1 gram / 1
        with self.assertRaises(ValueError):
            self.number == 1


class TestRatio(TestCase):

    point_five = Number(quantity=0.5)
    one = Number()
    two = Number(quantity=2)
    three = Number(quantity=3)
    four = Number(quantity=4)

    one_half = Ratio(numerator=one, denominator=two)
    three_fourths = Ratio(numerator=three, denominator=four)
    one_ratio = Ratio(numerator=one)
    three_halves = Ratio(numerator=three, denominator=two)
    two_ratio = Ratio(numerator=two, denominator=one)

    bromine_density = Ratio(Number(Units.gram, quantity=3.119), Number(Units.liter, Scale.milli))

    def test_evaluate(self):
        self.assertEqual(self.one_ratio.numerator, self.one)
        self.assertEqual(self.one_ratio.denominator, self.one)
        self.assertEqual(self.one_ratio.evaluate(), self.one)
        self.assertEqual(self.two_ratio.evaluate(), self.two)

    def test_reciprocal(self):
        self.assertEqual(self.two_ratio.reciprocal().numerator, self.one)
        self.assertEqual(self.two_ratio.reciprocal().denominator, self.two)
        self.assertEqual(self.two_ratio.reciprocal().evaluate(), self.point_five)

    def test___mul__(self):
        # Does commutivity hold?
        self.assertEqual(self.three_halves * self.one_half, self.three_fourths)
        self.assertEqual(self.one_half * self.three_halves, self.three_fourths)

        # How many grams of bromine are in 2 milliliters?
        two_milliliters_bromine = Number(Units.liter, Scale.milli, 2)
        answer = two_milliliters_bromine.as_ratio() * self.bromine_density
        self.assertEqual(answer.evaluate().value, 6.238) # Grams

    def test___eq__(self):
        self.assertEqual(self.one_half, self.point_five)
        with self.assertRaises(ValueError):
            self.one_half == 1/2

    def test___repr__(self):
        self.assertEqual(str(self.one_ratio), '<1.0 >')
        self.assertEqual(str(self.two_ratio), '<2 > / <1 >')
        self.assertEqual(str(self.two_ratio.evaluate()), '<2.0 >')

